package hamza.dali.flutter_osm_plugin;

import android.Manifest;
import android.app.Activity;
import android.app.Application;
import android.content.Context;
import android.content.pm.PackageManager;
import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.drawable.BitmapDrawable;
import android.graphics.drawable.Drawable;
import android.os.Build;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.LinearLayout;

import androidx.annotation.NonNull;
import androidx.core.content.ContextCompat;
import androidx.lifecycle.DefaultLifecycleObserver;
import androidx.lifecycle.Lifecycle;
import androidx.lifecycle.LifecycleOwner;

import org.osmdroid.config.Configuration;
import org.osmdroid.events.MapEventsReceiver;
import org.osmdroid.events.MapListener;
import org.osmdroid.events.ScrollEvent;
import org.osmdroid.events.ZoomEvent;
import org.osmdroid.tileprovider.MapTileProviderBase;
import org.osmdroid.tileprovider.tilesource.TileSourceFactory;
import org.osmdroid.util.GeoPoint;
import org.osmdroid.views.CustomZoomButtonsController;
import org.osmdroid.views.MapView;
import org.osmdroid.views.overlay.MapEventsOverlay;
import org.osmdroid.views.overlay.Marker;
import org.osmdroid.views.overlay.mylocation.GpsMyLocationProvider;
import org.osmdroid.views.overlay.mylocation.MyLocationNewOverlay;

import java.io.ByteArrayInputStream;
import java.nio.ByteBuffer;
import java.util.HashMap;
import java.util.concurrent.atomic.AtomicInteger;

import io.flutter.embedding.android.FlutterView;
import io.flutter.embedding.engine.FlutterEngine;
import io.flutter.embedding.engine.plugins.activity.ActivityAware;
import io.flutter.embedding.engine.plugins.activity.ActivityPluginBinding;
import io.flutter.plugin.common.BinaryMessenger;
import io.flutter.plugin.common.EventChannel;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.PluginRegistry;
import io.flutter.plugin.platform.PlatformView;

import static hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.CREATED;
import static hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.DESTROYED;
import static hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.PAUSED;
import static hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.RESUMED;
import static hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.STARTED;
import static hamza.dali.flutter_osm_plugin.FlutterOsmPlugin.STOPPED;
import static io.flutter.plugin.common.MethodChannel.Result;

public class FlutterOsmView implements
        Application.ActivityLifecycleCallbacks,
        DefaultLifecycleObserver,
        ActivityPluginBinding.OnSaveInstanceStateListener,
        PlatformView,
        EventChannel.StreamHandler,
        MethodChannel.MethodCallHandler {
    MapView map;
    private MyLocationNewOverlay locationNewOverlay;
    private Context context;
    private final MethodChannel methodChannel;
    private final int activityHashCode; // Do not use directly, use getActivityHashCode() instead to get correct hashCode for both v1 and v2 embedding.
    private final Lifecycle lifecycle;
    private final Application mApplication;
    private Activity mActivity;
    private final AtomicInteger activityState;
    private PluginRegistry.Registrar registrar;
    private Bitmap customMarkerIcon;
    MapEventsOverlay mapEventsOverlay;
    private final double defaultZoom=10.;

    public FlutterOsmView(Context ctx,
                          PluginRegistry.Registrar registrar,
                          BinaryMessenger binaryMessenger,
                          int id,
                          AtomicInteger activityState,
                          Application application,
                          Activity activity,
                          Lifecycle lifecycle,
                          int registrarActivityHashCode) {

        this.context = ctx;
        if (registrar != null)
            this.registrar = registrar;
        this.mActivity = activity;


        //LinearLayout view = (LinearLayout) getViewFromXML();
        Configuration.getInstance().load(context, PreferenceManager.getDefaultSharedPreferences(context));
        map = new MapView(context);
        map.setLayoutParams(new MapView.LayoutParams(new LinearLayout.LayoutParams(ViewGroup.LayoutParams.MATCH_PARENT, ViewGroup.LayoutParams.MATCH_PARENT)));
        map.setTilesScaledToDpi(true);
        //map.setZoomRounding(true);
        map.setTileSource(TileSourceFactory.MAPNIK);
        map.getZoomController().setVisibility(CustomZoomButtonsController.Visibility.NEVER);


        methodChannel = new MethodChannel(binaryMessenger, "plugins.dali.hamza/osmview_" + id);
        methodChannel.setMethodCallHandler(this);
        mApplication = application;
        this.lifecycle = lifecycle;
        this.activityHashCode = registrarActivityHashCode;
        this.activityState = activityState;
    }

    @Override
    public void onFlutterViewAttached(View flutterView) {

    }

    @Override
    public void onFlutterViewDetached() {
        map.onPause();
    }

    private View getViewFromXML() {
        return LayoutInflater.from(getApplication()).inflate(R.layout.osm_layout, null, false);
    }

    @Override
    public View getView() {

        return map;
    }

    private void setZoom(MethodCall methodCall, Result result) {
        double zoom = (double) methodCall.arguments;
        map.getController().setZoom(zoom);

        result.success(null);
    }

    private void initPosition(MethodCall methodCall, Result result) {

        HashMap<String, Double> args = (HashMap) methodCall.arguments;
        map.getOverlays().clear();
        GeoPoint geoPoint = new GeoPoint(args.get("lat"), args.get("lon"));
        addMarker(geoPoint,defaultZoom);
        result.success(null);
    }

    void addMarker(GeoPoint geoPoint,double zoom){
        Marker marker = new Marker(map);
        Drawable iconDrawable;
        if (customMarkerIcon != null) {
            iconDrawable = new BitmapDrawable(getActivity().getResources(), customMarkerIcon);
        } else {
            if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP) {
                iconDrawable = getActivity().getDrawable(R.drawable.ic_location_on_red_24dp);
            } else {
                iconDrawable = ContextCompat.getDrawable(getActivity(), R.drawable.ic_location_on_red_24dp);
            }
        }

        marker.setIcon(iconDrawable);

        marker.setPosition(geoPoint);
        map.getController().setZoom(10.);
        map.getController().animateTo(geoPoint);
        map.getOverlays().add(marker);
    }

    private void enableMyLocation(MethodCall methodCall, Result result) {

        map.getOverlays().clear();
        this.locationNewOverlay = new MyLocationNewOverlay(new GpsMyLocationProvider(getApplication()), map);
        this.locationNewOverlay.enableMyLocation();
        this.locationNewOverlay.runOnFirstFix(new Runnable() {
            @Override
            public void run() {
                if (getActivity() != null) {
                    getActivity().runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            GeoPoint geo = new GeoPoint(locationNewOverlay.getLastFix().getLatitude(), locationNewOverlay.getLastFix().getLongitude());
                            //map.getController().zoomToSpan(Math.abs(geo.getLatitude()),Math.abs(geo.getLongitude()));
                            map.getController().setZoom(15.);
                            map.getController().animateTo(geo);
                        }
                    });
                } else {
                    Log.d("mActivity ", "null");
                }
            }
        });

        map.getOverlays().add(this.locationNewOverlay);


        result.success(null);
    }

    @Override
    public void dispose() {

        methodChannel.setMethodCallHandler(null);
        getApplication().unregisterActivityLifecycleCallbacks(this);

    }

    @Override
    public void onMethodCall(MethodCall call, MethodChannel.Result result) {
        switch (call.method) {
            case "Zoom":
                setZoom(call, result);
                break;
            case "currentLocation":
                enableMyLocation(call, result);
                break;
            case "showZoomController":
                showZoomController(call, result);
                break;
            case "initPosition":
                initPosition(call, result);
                break;
            case "trackMe":
                enableTracking(call, result);
                break;
            case "user#position":
                userPosition(call, result);
                break;
            case "user#pickPosition":
                pickPosition(call, result);
                break;
            case "Road":
                result.notImplemented();
                break;
            case "marker#icon":
                changeIcon(call, result);
                break;
            default:
                result.notImplemented();
        }

    }
    private void pickPosition(MethodCall call,Result result){
        final Result _result = result;
        if(mapEventsOverlay == null){
            mapEventsOverlay =new MapEventsOverlay( new MapEventsReceiver() {
                @Override
                public boolean singleTapConfirmedHelper(GeoPoint p) {
                    return false;
                }

                @Override
                public boolean longPressHelper(GeoPoint p) {
                    HashMap<String, Double> Hashmap = new HashMap<>();
                    Hashmap.put("lat", p.getLatitude());
                    Hashmap.put("lon", p.getLongitude());
                    map.getOverlays().remove(0);
                    addMarker(p,15.);
                    mapEventsOverlay = null;
                    _result.success(Hashmap);
                    return false;
                }
            });
            map.getOverlays().add(0, mapEventsOverlay);
        }

    }
    private void changeIcon(MethodCall call, Result result) {
        try {
            byte[] bytes = (byte[]) call.arguments;
            Log.d("bytes", "" + call.arguments.toString());
            customMarkerIcon = BitmapFactory.decodeByteArray(bytes, 0, bytes.length);
            //customMarkerIcon.recycle();
            result.success(null);
        } catch (Exception e) {
            Log.d("err", e.getMessage());
            customMarkerIcon = null;
            result.error("500", "Cannot make markerIcon custom", "");
        }

    }

    private void userPosition(MethodCall call, Result result) {
        if (this.locationNewOverlay == null || !this.locationNewOverlay.isMyLocationEnabled()) {
            result.error("400", "current location is not enabled yet!", null);
        } else {
            if (this.locationNewOverlay.getLastFix() != null) {
                HashMap<String, Double> map = new HashMap<>();
                GeoPoint geo = new GeoPoint(locationNewOverlay.getLastFix().getLatitude(), locationNewOverlay.getLastFix().getLongitude());
                map.put("lat", geo.getLatitude());
                map.put("lon", geo.getLongitude());
                result.success(map);
            } else
                result.error("400", "location not available yet!", null);
        }


    }

    private void showZoomController(MethodCall call, Result result) {
        boolean showZoom = (boolean) call.arguments;
        map.getZoomController().setVisibility(showZoom ? CustomZoomButtonsController.Visibility.ALWAYS : CustomZoomButtonsController.Visibility.NEVER);
    }

    private void enableTracking(MethodCall call, Result result) {
        if (this.locationNewOverlay != null) {
            if (locationNewOverlay.isFollowLocationEnabled()) {
                locationNewOverlay.disableFollowLocation();
            } else {
                locationNewOverlay.enableFollowLocation();
            }
        }
    }

    private Activity getActivity() {
        if (registrar != null && registrar.activity() != null) {
            return registrar.activity();
        } else {
            return mActivity;
        }
    }

    private Application getApplication() {
        if (registrar != null && registrar.activity() != null) {
            return registrar.activity().getApplication();
        } else {
            return mApplication;
        }
    }

    @Override
    public void onActivityCreated(Activity activity, Bundle bundle) {

    }

    @Override
    public void onActivityStarted(Activity activity) {

    }

    @Override
    public void onActivityResumed(Activity activity) {
        map.onResume();
    }

    @Override
    public void onActivityPaused(Activity activity) {
        map.onPause();
    }

    @Override
    public void onActivityStopped(Activity activity) {

    }

    @Override
    public void onActivitySaveInstanceState(Activity activity, Bundle bundle) {

    }

    @Override
    public void onActivityDestroyed(Activity activity) {

    }

    @Override
    public void onSaveInstanceState(Bundle bundle) {

    }

    @Override
    public void onRestoreInstanceState(Bundle bundle) {

    }

    @Override
    public void onCreate(@NonNull LifecycleOwner owner) {

    }

    @Override
    public void onStart(@NonNull LifecycleOwner owner) {

    }

    @Override
    public void onResume(@NonNull LifecycleOwner owner) {
        map.onResume();
    }

    @Override
    public void onPause(@NonNull LifecycleOwner owner) {
        map.onPause();
    }

    @Override
    public void onStop(@NonNull LifecycleOwner owner) {

    }

    @Override
    public void onDestroy(@NonNull LifecycleOwner owner) {

    }

    void init() {
        switch (activityState.get()) {
            case STOPPED:
                break;
            case PAUSED:
                map.onPause();
                break;
            case RESUMED:
                map.onResume();
                break;
            case STARTED:
                break;
            case CREATED:
                break;
            case DESTROYED:
                // Nothing to do, the activity has been completely destroyed.
                break;
            default:
                throw new IllegalArgumentException(
                        "Cannot interpret " + activityState.get() + " as an activity state");
        }
        if (lifecycle != null) {
            lifecycle.addObserver(this);
        } else {
            getApplication().registerActivityLifecycleCallbacks(this);
        }

    }

    @Override
    public void onListen(Object arguments, EventChannel.EventSink events) {

    }

    @Override
    public void onCancel(Object arguments) {

    }
}
